!{\src2tex{textfont=tt}}
!!****f* ABINIT/newvtr3
!! NAME
!! newvtr3
!!
!! FUNCTION
!! Compute new first-order trial potential by mixing new and old values.
!! First, compute preconditioned residual first-order potential.
!! Then, call one of the self-consistency drivers, and  update vtrial.
!!
!! COPYRIGHT
!! Copyright (C) 1998-2009 ABINIT group (DCA, XG, MT)
!! This file is distributed under the terms of the
!! GNU General Public License, see ~abinit/COPYING
!! or http://www.gnu.org/copyleft/gpl.txt .
!!
!! INPUTS
!!  cplex: if 1, real space 1-order functions on FFT grid are REAL, if 2, COMPLEX
!!  dielar(7)=input parameters for dielectric matrix:
!!                diecut,dielng,diemac,diemix,diegap,dielam,diemixmag.
!!  dtset <type(dataset_type)>=all input variables in this dataset
!!   | isecur=level of security of the computation
!!   | mffmem=governs the number of FFT arrays which are fit in core memory
!!   |          it is either 1, in which case the array f_fftgr is used,
!!   |          or 0, in which case the array f_fftgr_disk is used
!!   | natom=number of atoms
!!   | nspden=number of spin-density components
!!   | paral_kgb=option for (kpt,g vectors,bands) parallelism
!!   | pawoptmix= - PAW only - 1 if the computed residuals include the PAW (rhoij) part
!!  etotal=the total energy obtained from the input vtrial
!!  ffttomix(nfft*(1-nfftmix/nfft))=Index of the points of the FFT (fine) grid on the grid used for mixing (coarse)
!!  filfft=name of _FFT file
!!  initialized= if 0 the initialization of the RF run is not yet finished
!!   iscf=( <= 0 =>non-SCF), >0 => SCF)
!!    iscf =1 => determination of the largest eigenvalue of the SCF cycle
!!    iscf =2 => SCF cycle, simple mixing
!!    iscf =3 => SCF cycle, Anderson mixing
!!    iscf =4 => SCF cycle, Anderson mixing (order 2)
!!    iscf =5 => SCF cycle, CG based on the minimization of the energy
!!    iscf =7 => SCF cycle, Pulay mixing
!!  ispmix=1 if mixing is done in real space, 2 if mixing is done in reciprocal space
!!  istep= number of the step in the SCF cycle
!!  i_rhor(n_index)=indices of the density in the array f_fftgr
!!  i_vresid(n_index)=indices of the potential residuals in the array f_fftgr
!!  i_vrespc(n_index)=indices of the preconditioned potential residuals in the array f_fftgr
!!  i_vtrial(n_index)=indices of the potential in the array f_fftgr
!!  mixtofft(nfftmix*(1-nfftmix/nfft))=Index of the points of the FFT grid used for mixing (coarse) on the FFT (fine) grid
!!  mpi_enreg=informations about MPI parallelization
!!  nfft=(effective) number of FFT grid points (for this processor)
!!  nfftmix=dimension of FFT grid used to mix the densities (used in PAW only)
!!  ngfft(18)=contain all needed information about 3D FFT, see ~abinit/doc/input_variables/vargs.htm#ngfft
!!  ngfftmix(18)=contain all needed information about 3D FFT, for the grid corresponding to nfftmix
!!  npawmix=-PAW only- number of spherical part elements to be mixed
!!  n_fftgr=third dimension of the array f_fftgr
!!  n_index=dimension for indices of potential/density (see i_vresid, ivrespc, i_vtrial...)
!!  qphon(3)=reduced coordinates for the phonon wavelength (needed if cplex==2).
!!  pawrhoij(natom) <type(pawrhoij_type)>= paw rhoij occupancies and related data
!!                                         Use here rhoij residuals (and gradients)
!!  rhor(cplex*nfft,nspden)=array for 1st-order electron density
!!    in electrons/bohr**3.
!!  rprimd(3,3)=dimensional primitive translations in real space (bohr)
!!  usepaw= 0 for non paw calculation; =1 for paw calculation
!!  vresid(cplex*nfft,nspden)=array for the residual of the potential
!!  xred(3,natom)=reduced dimensionless atomic coordinates
!!
!! OUTPUT
!!  dbl_nnsclo=1 if nnsclo has to be doubled to secure the convergence.
!!
!! SIDE EFFECTS
!!  f_fftgr(cplex*nfft,nspden,n_fftgr*mffmem)=different functions defined
!!    on the fft grid
!!   (see prcref, scfeig, scfopt, and scfcge for a detailed explanation).
!!   If mffmem=0, these data are kept on disk.
!!  vtrial(cplex*nfft,nspden)= at input, it is the "in" trial potential that gave vresid=(v_out-v_in)
!!       at output, it is an updated "mixed" trial potential
!!  ==== if usepaw==1
!!    f_paw(npawmix,n_fftgr*mffmem*usepaw)=different functions used for PAW
!!                                           (same as f_fftgr but for spherical part)
!!    pawrhoij(natom)%nrhoijsel,rhoijselect,rhoijp= several arrays
!!                containing new values of rhoij (augmentation occupancies)
!!
!! NOTES
!!  In case of PAW calculations:
!!    Computations are done either on the fine FFT grid or the coarse grid (depending on dtset%pawmixdg)
!!    All variables (nfft,ngfft,mgfft) refer to the fine FFT grid.
!!    All arrays (densities/potentials...) are computed on this fine FFT grid.
!!  ! Developpers have to be careful when introducing others arrays:
!!      they have to be stored on the fine FFT grid.
!!  In case of norm-conserving calculations the FFT grid is the usual FFT grid.
!!
!!  Subtility in PAW and non-collinear magnetism:
!!    Potentials are stored in (up-up,dn-dn,Re[up-dn],Im[up-dn]) format
!!    On-site occupancies (rhoij) are stored in (n,mx,my,mz)
!!    This is compatible provided that the mixing factors for n and m are identical
!!    and that the residual is not a combination of V_res and rhoij_res (pawoptmix=0).
!!
!! PARENTS
!!      scfcv3
!!
!! CHILDREN
!!      flush_unit,fourdp,leave_new,metric,moddiel,scfcge,scfopt,timab,wrtout
!!
!! SOURCE

#if defined HAVE_CONFIG_H
#include "config.h"
#endif

subroutine newvtr3(cplex,dbl_nnsclo,dielar,dtset,etotal,filfft,f_fftgr,f_paw,ffttomix,&
&          initialized,iscf,ispmix,istep,i_rhor,i_vresid,i_vrespc,i_vtrial,mixtofft,&
&          mpi_enreg,nfft,nfftmix,ngfft,ngfftmix,npawmix,n_fftgr,n_index,pawrhoij,&
&          qphon,rhor,rprimd,usepaw,vresid,vtrial,xred)

 use defs_basis
 use defs_datatypes
 use defs_abitypes
 use m_io_tools, only : flush_unit

!This section has been created automatically by the script Abilint (TD).
!Do not modify the following lines by hand.
 use interfaces_14_hidewrite
 use interfaces_16_hideleave
 use interfaces_18_timing
 use interfaces_42_geometry
 use interfaces_53_ffts
 use interfaces_67_common
!End of the abilint section

 implicit none

!Arguments-------------------------------
!scalars
 integer,intent(in) :: cplex,initialized,iscf,ispmix,istep,n_fftgr,n_index,nfft
 integer,intent(in) :: nfftmix,npawmix,usepaw
 integer,intent(out) :: dbl_nnsclo
 real(dp),intent(in) :: etotal
 character(len=fnlen),intent(in) :: filfft
 type(MPI_type),intent(inout) :: mpi_enreg
 type(dataset_type),intent(in) :: dtset
!arrays
 integer,intent(in) :: ffttomix(nfft*(1-nfftmix/nfft))
 integer,intent(in) :: mixtofft(nfftmix*(1-nfftmix/nfft)),ngfft(18)
 integer,intent(in) :: ngfftmix(18)
 integer,intent(inout) :: i_rhor(n_index),i_vresid(n_index),i_vrespc(n_index)
 integer,intent(inout) :: i_vtrial(n_index)
 real(dp),intent(in) :: dielar(7),qphon(3),rhor(cplex*nfft,dtset%nspden)
 real(dp),intent(in) :: rprimd(3,3)
 real(dp),intent(inout) :: f_fftgr(cplex*nfft,dtset%nspden,n_fftgr*dtset%mffmem)
 real(dp),intent(inout) :: f_paw(npawmix,n_fftgr*dtset%mffmem*usepaw)
 real(dp),intent(inout) :: vresid(cplex*nfft,dtset%nspden)
 real(dp),intent(inout) :: vtrial(cplex*nfft,dtset%nspden),xred(3,dtset%natom)
 type(pawrhoij_type),intent(inout) :: pawrhoij(dtset%natom*usepaw)

!Local variables-------------------------------
!scalars
 integer :: cplex_mix,cplex_rhoij,dplex,i_vresid1,i_vrespc1,iatom,ifft,indx
 integer :: irhoij,ispden,jfft,jrhoij,klmn,kmix,moved_atm_inside,nfftot,nselect
 integer :: response
 real(dp) :: fact,mixfac,mixfac_eff,mixfacmag,ucvol
 character(len=500) :: message
!arrays
 real(dp) :: gmet(3,3),gprimd(3,3),rmet(3,3),tsec(2)
 real(dp),allocatable :: dtn_pc(:,:),f_atm(:,:,:),f_fftgr_disk(:,:,:)
 real(dp),allocatable :: f_paw_disk(:,:),rhoijrespc(:),rhoijtmp(:,:),vpaw(:)
 real(dp),allocatable :: vresid0(:,:),vrespc(:,:),vreswk(:,:),vtrial0(:,:)
 real(dp),allocatable :: vtrialg(:,:,:)

! *************************************************************************

#if defined DEBUG_MODE
 write(message,'(a)')' newvtr3 : enter '
 call wrtout(std_out,message,'COLL')
 call flush(std_out)
#endif

!Compatibility tests
 if(usepaw==1) then
  if(dtset%nspden==4.and.dtset%pawoptmix==1) then
   write(message, '(a,a,a,a)' )ch10,&
&   ' newvtr : ERROR -',ch10,&
&   '  pawoptmix=1 is not compatible with nspden=4 !'
   call wrtout(6,message,'COLL')
   call leave_new('PERS')
  end if
  if (pawrhoij(1)%cplex<cplex) then
   write(message, '(a,a,a,a)' )ch10,&
&   ' newvtr3 : BUG -',ch10,&
&   '  pawrhoij()%cplex must be >=cplex !'
   call wrtout(6,message,'COLL')
   call leave_new('COLL')
  end if
 end if

 call timab(158,1,tsec)

 nfftot=ngfft(1)*ngfft(2)*ngfft(3)
 cplex_mix=max(cplex,ispmix)
 if (usepaw==1) cplex_rhoij=pawrhoij(1)%cplex

!Compute different geometric tensor, as well as ucvol, from rprimd
 call metric(gmet,gprimd,-1,rmet,rprimd,ucvol)
 moved_atm_inside=0

!Select components of potential to be mixed
 allocate(vtrial0(cplex_mix*nfftmix,dtset%nspden),vresid0(cplex_mix*nfftmix,dtset%nspden))
 if (ispmix==1.and.nfft==nfftmix) then
  vtrial0=vtrial;vresid0=vresid
 else if (nfft==nfftmix) then
  do ispden=1,dtset%nspden
   call fourdp(cplex,vtrial0(:,ispden),vtrial(:,ispden),-1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
   call fourdp(cplex,vresid0(:,ispden),vresid(:,ispden),-1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
  end do
 else
  allocate(vtrialg(2,nfft,dtset%nspden),vreswk(2,nfft))
  do ispden=1,dtset%nspden
   fact=dielar(4);if (ispden>1) fact=dielar(7)
   call fourdp(cplex,vtrialg(:,:,ispden),vtrial(:,ispden),-1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
   call fourdp(cplex,vreswk,vresid(:,ispden),-1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
   do ifft=1,nfft
    if (ffttomix(ifft)>0) then
     jfft=2*ffttomix(ifft)
     vtrial0(jfft-1,ispden)=vtrialg(1,ifft,ispden)
     vtrial0(jfft  ,ispden)=vtrialg(2,ifft,ispden)
     vresid0(jfft-1,ispden)=vreswk(1,ifft)
     vresid0(jfft  ,ispden)=vreswk(2,ifft)
    else
     vtrialg(:,ifft,ispden)=vtrialg(:,ifft,ispden)+fact*vreswk(:,ifft)
    end if
   end do
  end do
  deallocate(vreswk)
 end if

!Precondition the potential residual:
!Use a model dielectric function preconditioning, or simple mixing
 allocate(vrespc(cplex_mix*nfft,dtset%nspden))
 call moddiel(cplex_mix,dielar,mpi_enreg,nfftmix,ngfftmix,dtset%nspden,ispmix,0,&
& dtset%paral_kgb,qphon,rprimd,vresid0,vrespc)

!PAW only : precondition the rhoij quantities (augmentation occupancies) residuals.
!Use a simple preconditionning with the same mixing factor
!as the model dielectric function.
 if (usepaw==1) then
  allocate(rhoijrespc(npawmix))
  mixfac=dielar(4);mixfacmag=abs(dielar(7))
  if (cplex_rhoij==1) then
   indx=0
   do iatom=1,dtset%natom
    do ispden=1,pawrhoij(iatom)%nspden
     mixfac_eff=mixfac;if (ispden>1) mixfac_eff=mixfacmag
     do kmix=1,pawrhoij(iatom)%lmnmix_sz
      indx=indx+1;klmn=pawrhoij(iatom)%kpawmix(kmix)
      rhoijrespc(indx)=mixfac_eff*pawrhoij(iatom)%rhoijres(klmn,ispden)
     end do
    end do
   end do
  else
   indx=-1
   do iatom=1,dtset%natom
    do ispden=1,pawrhoij(iatom)%nspden
     mixfac_eff=mixfac;if (ispden>1) mixfac_eff=mixfacmag
     do kmix=1,pawrhoij(iatom)%lmnmix_sz
      indx=indx+2;klmn=2*pawrhoij(iatom)%kpawmix(kmix)-1
      rhoijrespc(indx:indx+1)=mixfac_eff*pawrhoij(iatom)%rhoijres(klmn:klmn+1,ispden)
     end do
    end do
   end do
  end if
 end if

!------Compute new vtrial

 i_vresid1=i_vresid(1)
 i_vrespc1=i_vrespc(1)

!Either use the array f_fftgr or the array f_fftgr_disk
 if(dtset%mffmem==1)then
  f_fftgr(:,:,i_vresid1)=vresid0(:,:)
  f_fftgr(:,:,i_vrespc1)=vrespc(:,:)
 else
! In this case, must first allocate f_fftgr_disk, then take data
! from disk and existing arrays.
  allocate(f_fftgr_disk(cplex_mix*nfftmix,dtset%nspden,n_fftgr))
  if(istep/=1)then
   call timab(83,1,tsec)
   open(unit=tmp_unit,file=filfft,form='unformatted',status='old')
   rewind(tmp_unit)
   read(tmp_unit)f_fftgr_disk
   close(unit=tmp_unit)
   call timab(83,2,tsec)
  end if
  f_fftgr_disk(:,:,i_vresid1)=vresid0(:,:)
  f_fftgr_disk(:,:,i_vrespc1)=vrespc(:,:)
 end if
 deallocate(vrespc)

!PAW: either use the array f_paw or the array f_paw_disk
 if (usepaw==1) then
  allocate(vpaw(npawmix),rhoijrespc(npawmix))
  dplex=cplex_rhoij-1
  if(dtset%mffmem==1)then
   indx=-dplex
   do iatom=1,dtset%natom
    do ispden=1,pawrhoij(iatom)%nspden
     allocate(rhoijtmp(cplex_rhoij*pawrhoij(iatom)%lmn2_size,1));rhoijtmp=zero
     jrhoij=1
     do irhoij=1,pawrhoij(iatom)%nrhoijsel
      klmn=cplex_rhoij*pawrhoij(iatom)%rhoijselect(irhoij)-dplex
      rhoijtmp(klmn:klmn+dplex,1)=pawrhoij(iatom)%rhoijp(jrhoij:jrhoij+dplex,ispden)
      jrhoij=jrhoij+cplex_rhoij
     end do
     do kmix=1,pawrhoij(iatom)%lmnmix_sz
      indx=indx+cplex_rhoij;klmn=cplex_rhoij*pawrhoij(iatom)%kpawmix(kmix)-dplex
      vpaw(indx:indx+dplex)=rhoijtmp(klmn:klmn+dplex,1)-pawrhoij(iatom)%rhoijres(klmn:klmn+dplex,ispden)
      f_paw(indx:indx+dplex,i_vresid1)=pawrhoij(iatom)%rhoijres(klmn:klmn+dplex,ispden)
      f_paw(indx:indx+dplex,i_vrespc1)=rhoijrespc(indx:indx+dplex)
     end do
     deallocate(rhoijtmp)
    end do
   end do
  else
   allocate(f_paw_disk(npawmix,n_fftgr))
   if(istep/=1)then
    call timab(83,1,tsec)
    read(tmp_unit)f_paw_disk
    close(unit=tmp_unit)
    call timab(83,2,tsec)
   end if
   indx=1-cplex_rhoij
   do iatom=1,dtset%natom
    do ispden=1,pawrhoij(iatom)%nspden
     allocate(rhoijtmp(cplex_rhoij*pawrhoij(iatom)%lmn2_size,1));rhoijtmp=zero
     jrhoij=1
     do irhoij=1,pawrhoij(iatom)%nrhoijsel
      klmn=cplex_rhoij*pawrhoij(iatom)%rhoijselect(irhoij)-dplex
      rhoijtmp(klmn:klmn+dplex,1)=pawrhoij(iatom)%rhoijp(jrhoij:jrhoij+dplex,ispden)
      jrhoij=jrhoij+cplex_rhoij
     end do
     do kmix=1,pawrhoij(iatom)%lmnmix_sz
      indx=indx+cplex_rhoij;klmn=cplex_rhoij*pawrhoij(iatom)%kpawmix(kmix)-dplex
      vpaw(indx:indx+dplex)=rhoijtmp(klmn:klmn+dplex,1)-pawrhoij(iatom)%rhoijres(klmn:klmn+dplex,ispden)
      f_paw_disk(indx:indx+dplex,i_vresid1)=pawrhoij(iatom)%rhoijres(klmn:klmn+dplex,ispden)
      f_paw_disk(indx:indx+dplex,i_vrespc1)=rhoijrespc(indx:indx+dplex)
     end do
     deallocate(rhoijtmp)
    end do
   end do
  end if
 end if

!Unlike for GS, no need to modify the mean of vtrial

!Do here the mixing of the potential
 if(iscf==2 .or. iscf==3 .or. iscf==7)then

! Optimize next vtrial using different algorithms,
! as determined by the variable iscf
  allocate(dtn_pc(3,dtset%natom))
  if(dtset%mffmem==1)then
   call scfopt(cplex_mix,dtn_pc,f_fftgr,f_paw,iscf,istep,i_vrespc,i_vtrial,&
&   moved_atm_inside,mpi_enreg,dtset%natom,nfftmix,npawmix,dtset%nspden,n_fftgr,n_index,&
&   0,dtset%pawoptmix,usepaw,vpaw,vtrial0,xred)
  else
   call scfopt(cplex_mix,dtn_pc,f_fftgr_disk,f_paw_disk,iscf,istep,i_vrespc,i_vtrial,&
&   moved_atm_inside,mpi_enreg,dtset%natom,nfftmix,npawmix,dtset%nspden,n_fftgr,n_index,&
&   0,dtset%pawoptmix,usepaw,vpaw,vtrial0,xred)
  end if
  deallocate(dtn_pc)

! PAW: restore rhoij from compact storage
  if (usepaw==1) then
   indx=-dplex
   do iatom=1,dtset%natom
    allocate(rhoijtmp(cplex_rhoij*pawrhoij(iatom)%lmn2_size,pawrhoij(iatom)%nspden));rhoijtmp=zero
    if (pawrhoij(iatom)%lmnmix_sz<pawrhoij(iatom)%lmn2_size) then
     do ispden=1,pawrhoij(iatom)%nspden
      jrhoij=1
      do irhoij=1,pawrhoij(iatom)%nrhoijsel
       klmn=cplex_rhoij*pawrhoij(iatom)%rhoijselect(irhoij)-dplex
       rhoijtmp(klmn:klmn+dplex,ispden)=pawrhoij(iatom)%rhoijp(jrhoij:jrhoij+dplex,ispden)
       jrhoij=jrhoij+cplex_rhoij
      end do
     end do
    end if
    do ispden=1,pawrhoij(iatom)%nspden
     do kmix=1,pawrhoij(iatom)%lmnmix_sz
      indx=indx+cplex_rhoij;klmn=cplex_rhoij*pawrhoij(iatom)%kpawmix(kmix)-dplex
      rhoijtmp(klmn:klmn+dplex,ispden)=vpaw(indx:indx+dplex)
     end do
    end do
    nselect=0
    do klmn=1,pawrhoij(iatom)%lmn2_size
     if (any(abs(rhoijtmp(cplex_rhoij*klmn-dplex:cplex_rhoij*klmn,:))>tol10)) then
      nselect=nselect+1
      pawrhoij(iatom)%rhoijselect(nselect)=klmn
      do ispden=1,pawrhoij(iatom)%nspden
       pawrhoij(iatom)%rhoijp(cplex_rhoij*nselect-dplex:cplex_rhoij*nselect,ispden)=&
&       rhoijtmp(cplex_rhoij*klmn-dplex:cplex_rhoij*klmn,ispden)
      end do
     end if
    end do
    pawrhoij(iatom)%nrhoijsel=nselect
    deallocate(rhoijtmp)
   end do
   deallocate(vpaw)
  end if

 else if(iscf==5 .or. iscf==6)then

  if(ispmix/=1) then
   write(message, '(a,a,a,a)' )ch10,&
&   ' newvtr3 : ERROR -',ch10,&
&   '  Mixing on reciprocal space not allowed with iscf=5 or 6.'
   call wrtout(6,message,'COLL')
   call leave_new('PERS')
  end if
! Optimize next vtrial using an algorithm based
! on the conjugate gradient minimization of etotal
  response=1
  allocate(f_atm(3,dtset%natom,n_fftgr),dtn_pc(3,dtset%natom))
  if(dtset%mffmem==1)then
   call scfcge(cplex,dbl_nnsclo,dtn_pc,etotal,f_atm,&
&   f_fftgr,initialized,iscf,dtset%isecur,istep,&
&   i_rhor,i_vresid,i_vrespc,moved_atm_inside,mpi_enreg,&
&   dtset%natom,nfft,nfftot,dtset%nspden,n_fftgr,n_index,0,&
&   response,rhor,ucvol,vtrial0,xred)
  else
   call scfcge(cplex,dbl_nnsclo,dtn_pc,etotal,f_atm,&
&   f_fftgr_disk,initialized,iscf,dtset%isecur,istep,&
&   i_rhor,i_vresid,i_vrespc,moved_atm_inside,mpi_enreg,&
&   dtset%natom,nfft,nfftot,dtset%nspden,n_fftgr,n_index,0,&
&   response,rhor,ucvol,vtrial0,xred)
  end if
  deallocate(f_atm,dtn_pc)
! PAW: apply a simple mixing to rhoij (this is temporary)
  if (usepaw==1) then
   indx=1-cplex_rhoij
   do iatom=1,dtset%natom
    allocate(rhoijtmp(cplex_rhoij*pawrhoij(iatom)%lmn2_size,pawrhoij(iatom)%nspden));rhoijtmp=zero
    if (pawrhoij(iatom)%lmnmix_sz<pawrhoij(iatom)%lmn2_size) then
     do ispden=1,pawrhoij(iatom)%nspden
      do kmix=1,pawrhoij(iatom)%lmnmix_sz
       indx=indx+cplex_rhoij;klmn=cplex_rhoij*pawrhoij(iatom)%kpawmix(kmix)-dplex
       rhoijtmp(klmn:klmn+dplex,ispden)=rhoijrespc(indx:indx+dplex) &
&       -pawrhoij(iatom)%rhoijres(klmn:klmn+dplex,ispden)
      end do
     end do
    end if
    do ispden=1,pawrhoij(iatom)%nspden
     jrhoij=1
     do irhoij=1,pawrhoij(iatom)%nrhoijsel
      klmn=cplex_rhoij*pawrhoij(iatom)%rhoijselect(irhoij)-dplex
      rhoijtmp(klmn:klmn+dplex,ispden)=rhoijtmp(klmn:klmn+dplex,ispden) &
&      +pawrhoij(iatom)%rhoijp(jrhoij:jrhoij+dplex,ispden)
      jrhoij=jrhoij+cplex_rhoij
     end do
    end do
    nselect=0
    do klmn=1,pawrhoij(iatom)%lmn2_size
     if (any(abs(rhoijtmp(cplex_rhoij*klmn-dplex:cplex_rhoij*klmn,:))>tol10)) then
      nselect=nselect+1
      pawrhoij(iatom)%rhoijselect(nselect)=klmn
      do ispden=1,pawrhoij(iatom)%nspden
       pawrhoij(iatom)%rhoijp(cplex_rhoij*nselect-dplex:cplex_rhoij*nselect,ispden)=&
&       rhoijtmp(cplex_rhoij*klmn-dplex:cplex_rhoij*klmn,ispden)
      end do
     end if
    end do
    pawrhoij(iatom)%nrhoijsel=nselect
    deallocate(rhoijtmp)
   end do
  end if

 else

  write(message, '(a,a,a,a,i5,a)' ) ch10,&
&  ' newvtr3 : BUG -',ch10,&
&  '  Invalid option: iscf =',iscf,'.'
  call wrtout(06,message,'COLL')
  call leave_new('COLL')

 end if

 if (usepaw==1) deallocate(rhoijrespc)

!Eventually write the data on disk and deallocate f_fftgr_disk
 if (dtset%mffmem==0)then
  call timab(83,1,tsec)
  open(unit=tmp_unit,file=filfft,form='unformatted',status='unknown')
  rewind(tmp_unit)
  write(tmp_unit)f_fftgr_disk
  if (usepaw==1) write(tmp_unit)f_paw_disk
  close(unit=tmp_unit)
  deallocate(f_fftgr_disk);if (usepaw==1) deallocate(f_paw_disk)
  call timab(83,2,tsec)
 end if

!Restore potential
 if (ispmix==1.and.nfft==nfftmix) then
  vtrial=vtrial0
 else if (nfft==nfftmix) then
  do ispden=1,dtset%nspden
   call fourdp(cplex,vtrial0(:,ispden),vtrial(:,ispden),+1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
  end do
 else
  do ispden=1,dtset%nspden
   do ifft=1,nfftmix
    jfft=mixtofft(ifft)
    vtrialg(1,jfft,ispden)=vtrial0(2*ifft-1,ispden)
    vtrialg(2,jfft,ispden)=vtrial0(2*ifft  ,ispden)
   end do
   call fourdp(cplex,vtrialg(:,:,ispden),vtrial(:,ispden),+1,mpi_enreg,nfft,ngfft,dtset%paral_kgb,0)
  end do
  deallocate(vtrialg)
 end if
 deallocate(vtrial0)

 call timab(158,2,tsec)

#if defined DEBUG_MODE
 write(message,'(a)')' newvtr3 : exit '
 call wrtout(std_out,message,'COLL')
 call flush(std_out)
#endif

end subroutine newvtr3
!!***
